'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

function _interopDefault (ex) { return (ex && (typeof ex === 'object') && 'default' in ex) ? ex['default'] : ex; }

var fs = require('fs');
var dayjs = _interopDefault(require('dayjs'));
var util = _interopDefault(require('util'));
var chalk = _interopDefault(require('chalk'));

var Types = {
  // Level 0
  fatal: {
    level: 0
  },
  error: {
    level: 0
  },
  // Level 1
  warn: {
    level: 1
  },
  // Level 2
  log: {
    level: 2
  },
  // Level 3
  info: {
    level: 3
  },
  success: {
    level: 3
  },
  // Level 4
  debug: {
    level: 4
  },
  // Level 5
  trace: {
    level: 5
  },
  // Silent
  silent: {
    level: Infinity
  }
};

function isPlainObject(obj) {
  return Object.prototype.toString.call(obj) === '[object Object]';
}
function isLogObj(arg) {
  // Should be plain object
  // Also contains either message or args field
  return isPlainObject(arg) && Boolean(arg.message || arg.args);
}

let paused = false;
const queue = [];
class Consola {
  constructor(options = {}) {
    this._reporters = options.reporters || [];
    this._types = options.types || Types;
    this._level = options.level != null ? options.level : 3;
    this._defaults = options.defaults || {};
    this._async = typeof options.async !== 'undefined' ? options.async : null;

    if (typeof process !== 'undefined') {
      this._stdout = options.stdout || process.stdout;
      this._stderr = options.stdout || process.stderr;
    } // Create logger functions for current instance


    for (const type in this._types) {
      this[type] = this._wrapLogFn(Object.assign({
        type
      }, this._types[type], this._defaults));
    }
  }

  get level() {
    return this._level;
  }

  set level(newLevel) {
    // Ensure that newLevel does not exceeds type level boundaries
    let min = 0;
    let max = 0;

    for (const typeName in this._types) {
      const type = this._types[typeName];

      if (type.level > max) {
        max = type.level;
      } else if (type.level < min) {
        min = type.level;
      }
    } // Set level


    this._level = Math.min(max, Math.max(min, newLevel));
  }

  create(options) {
    return new Consola(Object.assign({
      reporters: this._reporters,
      level: this._level,
      types: this._types,
      defaults: this._defaults,
      stdout: this._stdout,
      stderr: this._stderr
    }, options));
  }

  addReporter(reporter) {
    this._reporters.push(reporter);

    return this;
  }

  removeReporter(reporter) {
    if (reporter) {
      const i = this._reporters.indexOf(reporter);

      if (i >= 0) {
        return this._reporters.splice(i, 1);
      }
    } else {
      this._reporters.splice(0);
    }

    return this;
  }

  setReporters(reporters) {
    this._reporters = Array.isArray(reporters) ? reporters : [reporters];
  }

  withDefaults(defaults) {
    return this.create({
      defaults: Object.assign({}, this._defaults, defaults)
    });
  }

  withTag(tag) {
    return this.withDefaults({
      tag: this._defaults.tag ? this._defaults.tag + ':' + tag : tag
    });
  }

  wrapAll() {
    this.wrapConsole();
    this.wrapStd();
  }

  restoreAll() {
    this.restoreConsole();
    this.restoreStd();
  }

  wrapConsole() {
    for (const type in this._types) {
      // Backup original value
      if (!console['__' + type]) {
        // eslint-disable-line no-console
        console['__' + type] = console[type]; // eslint-disable-line no-console
      } // Override


      console[type] = this[type]; // eslint-disable-line no-console
    }
  }

  restoreConsole() {
    for (const type in this._types) {
      // Restore if backup is available
      if (console['__' + type]) {
        // eslint-disable-line no-console
        console[type] = console['__' + type]; // eslint-disable-line no-console

        delete console['__' + type]; // eslint-disable-line no-console
      }
    }
  }

  wrapStd() {
    this._wrapStream(this._stdout, 'log');

    this._wrapStream(this._stderr, 'error');
  }

  _wrapStream(stream, type) {
    if (!stream) {
      return;
    } // Backup original value


    if (!stream.__write) {
      stream.__write = stream.write;
    } // Override


    stream.write = data => {
      this[type](String(data).trim());
    };
  }

  restoreStd() {
    this._restoreStream(this._stdout);

    this._restoreStream(this._stderr);
  }

  _restoreStream(stream) {
    if (!stream) {
      return;
    }

    if (stream.__write) {
      stream.write = stream.__write;
      delete stream.__write;
    }
  }

  pause() {
    paused = true;
  }

  resume() {
    paused = false; // Process queue

    const _queue = queue.splice(0);

    for (const item of _queue) {
      item[0]._logFn(item[1], item[2]);
    }
  }

  _wrapLogFn(defaults) {
    function logFn() {
      if (paused) {
        queue.push([this, defaults, arguments]);
        return;
      }

      return this._logFn(defaults, arguments);
    }

    return logFn.bind(this);
  }

  _logFn(defaults, args) {
    if (defaults.level > this._level) {
      return this._async ? Promise.resolve(false) : false;
    } // Construct a new log object


    const logObj = Object.assign({
      date: new Date(),
      args: []
    }, defaults); // Consume arguments

    if (args.length === 1 && isLogObj(args[0])) {
      Object.assign(logObj, args[0]);
    } else {
      logObj.args = Array.from(args);
    } // Addtional message


    if (logObj.additional) {
      logObj.args.push.apply(logObj.args, logObj.additional.split('\n'));
      delete logObj.additional;
    } // Log


    if (this._async) {
      return this._logAsync(logObj);
    } else {
      this._log(logObj);
    }
  }

  _log(logObj) {
    for (const reporter of this._reporters) {
      reporter.log(logObj, {
        async: false,
        stdout: this._stdout,
        stderr: this._stderr
      });
    }
  }

  _logAsync(logObj) {
    return Promise.all(this._reporters.map(reporter => reporter.log(logObj, {
      async: true,
      stdout: this._stdout,
      stderr: this._stderr
    })));
  }

}

function assignGlobalReference(newInstance, referenceKey) {
  if (!newInstance.constructor || global[referenceKey] && !global[referenceKey].constructor) {
    throw new Error('Assigning to global reference is only supported for class instances');
  } else if (newInstance.constructor && !global[referenceKey]) {
    global[referenceKey] = newInstance;
  } else if (!(newInstance instanceof global[referenceKey].constructor || global[referenceKey] instanceof newInstance.constructor)) {
    throw new Error(`Not a ${global[referenceKey].constructor.name} instance`);
  }

  const oldInstance = Object.create(global[referenceKey]);

  for (let prop in global[referenceKey]) {
    oldInstance[prop] = global[referenceKey][prop];
    delete global[referenceKey][prop];
  }

  for (let prop of Object.getOwnPropertySymbols(global[referenceKey])) {
    oldInstance[prop] = global[referenceKey][prop];
    delete global[referenceKey][prop];
  }

  for (let prop in newInstance) {
    global[referenceKey][prop] = newInstance[prop];
  }

  for (let prop of Object.getOwnPropertySymbols(newInstance)) {
    global[referenceKey][prop] = newInstance[prop];
  }

  return oldInstance;
}
function assignGlobalConsola(newConsola) {
  return assignGlobalReference(newConsola, 'consola');
}

function parseStack(stack) {
  let lines = stack.split('\n').map(l => l.trim().replace(/^at /, ''));

  if (lines[0].indexOf('Error: ') === 0) {
    lines = lines.splice(1);
  }

  return lines;
}

function writeStream(data, stream, mode = 'default') {
  const write = stream.__write || stream.write;

  switch (mode) {
    case 'async':
      return new Promise(resolve => {
        if (write.call(stream, data) === true) {
          resolve();
        } else {
          stream.once('drain', () => {
            resolve();
          });
        }
      });

    case 'sync':
      return fs.writeSync(stream.fd, data);

    default:
      return write.call(stream, data);
  }
}

function formatDate(timeFormat, date) {
  return dayjs(date).format(timeFormat);
}

function leftAlign(str, len, space = ' ') {
  let _str = '';

  for (let i = 0; i < len; i++) {
    _str += i < str.length ? str[i] : space;
  }

  return _str;
}

const DEFAULTS = {
  dateFormat: 'HH:mm:ss'
};
class BasicReporter {
  constructor(options) {
    this.options = Object.assign({}, DEFAULTS, options);
  }

  formatStack(stack) {
    return ' at ' + parseStack(stack).join('\n at ');
  }

  formatArg(arg) {
    if (isPlainObject(arg)) {
      return JSON.stringify(arg, null, 2);
    }

    if (util.formatWithOptions) {
      // Node >= 10
      return util.formatWithOptions({
        colors: true
      }, arg);
    } else {
      return util.format(arg);
    }
  }

  formatDate(date) {
    return formatDate(this.options.dateFormat, date);
  }

  getFields(logObj) {
    let message = logObj.message || '';
    let type = logObj.type || '';
    let tag = logObj.tag || '';
    let date = logObj.date || new Date(); // Format args

    const args = logObj.args.map(arg => {
      // error-like argument
      if (arg.stack) {
        if (!message.length && arg.message) {
          message = arg.message;
        }

        if (!type.length) {
          type = 'error';
        }

        if (!tag.length && arg.code) {
          tag = arg.code;
        }

        return this.formatStack(arg.stack);
      } // General argument


      return this.formatArg(arg);
    }); // If no message is provided, assume args[0] as message

    if (!message.length && args.length) {
      message = args.shift();
    } // Concert other args to additional


    const additional = args.join('\n');
    return {
      additional,
      date,
      message,
      tag,
      type
    };
  }

  formatLogObj(logObj) {
    const fields = this.getFields(logObj);
    const date = this.formatDate(fields.date);
    const type = leftAlign(fields.type.toUpperCase(), 7);
    return [`[${date}]`, `[${fields.tag}]`, `[${type}]`, `${fields.message}`, `${fields.additional ? '\n' + fields.additional : ''}`];
  }

  log(logObj, {
    async,
    stdout,
    stderr
  } = {}) {
    const line = this.formatLogObj(logObj).filter(x => x && x.length && x !== '[]').join(' ') + '\n';
    return writeStream(line, logObj.level < 2 ? stderr : stdout, async ? 'async' : 'default');
  }

}

const TYPE_NAME_MAP = {
  error: 'ERR!',
  fatal: 'DIE!',
  warn: 'WARN',
  success: 'OK',
  debug: '...',
  trace: '...',
  log: ''
};
const TYPE_COLOR_MAP = {
  'info': 'cyan'
};
const LEVEL_COLOR_MAP = {
  0: 'red',
  1: 'yellow',
  2: 'white',
  3: 'green'
};

class BrowserReporter {
  constructor(options) {
    this.options = Object.assign({}, options);
  }

  log(logObj) {
    // consoleLogFn
    let consoleLogFn = console[logObj.type]; // eslint-disable-line no-console

    if (!consoleLogFn) {
      consoleLogFn = console[logObj.level < 2 ? 'error' : 'log']; // eslint-disable-line no-console
    } // Type


    const type = leftAlign((TYPE_NAME_MAP[logObj.type] || logObj.type).toUpperCase(), 4); // Styles

    const color = TYPE_COLOR_MAP[logObj.type] || LEVEL_COLOR_MAP[logObj.level];
    const styleColor = `color: ${color}; background-color: inherit;`;
    const styleInherit = `color: inherit; background-color: inherit;`;
    const styleAdditional = `color: ${logObj.additionalColor || 'grey'}; background-color: inherit;`; // Date

    const date = new Date(logObj.date).toLocaleTimeString(); // Message

    const message = logObj.message ? logObj.message : // If no message is provided, assume args[0] as message
    logObj.args.length ? logObj.args.shift() : ''; // Additional

    const additonal = logObj.args.length ? '\n' + logObj.args.join() : ''; // Log to the console

    const args = `%c${type} %c${message} %c${additonal ? additonal + ' ' : ''}(${date})`;
    consoleLogFn(args, styleColor, styleInherit, styleAdditional);
  }

}

const _colorCache = {};
function chalkColor(name) {
  let color = _colorCache[name];

  if (color) {
    return color;
  }

  if (name[0] === '#') {
    color = chalk.hex(name);
  } else {
    color = chalk[name] || chalk.keyword(name);
  }

  _colorCache[name] = color;
  return color;
}

const DEFAULTS$1 = {
  secondaryColor: 'grey',
  tagColor: 'magenta',
  dateFormat: 'HH:mm'
};
class FancyReporter extends BasicReporter {
  constructor(options) {
    super(Object.assign({}, DEFAULTS$1, options));
  }

  formatStack(stack) {
    return ' at ' + parseStack(stack).join(' ↲\n at ');
  }

  typeColor(type, level) {
    return chalkColor(TYPE_COLOR_MAP[type] || LEVEL_COLOR_MAP[level] || this.options.secondaryColor);
  }

  formatType(type, typeColor) {
    if (TYPE_NAME_MAP[type] === '') {
      return '';
    }

    return typeColor(leftAlign((TYPE_NAME_MAP[type] || type).toUpperCase(), 4));
  }

  formatTag(tag) {
    return chalkColor(this.options.tagColor)(tag);
  }

  secondaryColor(str, additionalColor) {
    return chalkColor(additionalColor || this.options.secondaryColor)(str);
  }

  formatLogObj(logObj) {
    const fields = this.getFields(logObj);
    const typeColor = this.typeColor(fields.type, logObj.level);
    const date = this.secondaryColor(`(${this.formatDate(fields.date)})`);
    const type = this.formatType(fields.type, typeColor);
    const tag = this.formatTag(fields.tag);
    const message = fields.message;
    const additional = fields.additional.length ? this.secondaryColor('\n' + fields.additional, logObj.additionalColor) : '';
    return [`${tag}`, `${type}`, `${message}`, `${date}`, `${additional}`];
  }

}

class JSONReporter {
  constructor({
    stream
  }) {
    this.stream = stream || process.stdout;
  }

  log(logObj) {
    this.stream.write(JSON.stringify(logObj) + '\n');
  }

}

// This reporter is compatible with Winston 3
// https://github.com/winstonjs/winston
class WinstonReporter {
  constructor(logger) {
    if (logger && logger.log) {
      this.logger = logger;
    } else {
      const winston = require('winston');

      this.logger = winston.createLogger(Object.assign({
        level: 'info',
        format: winston.format.simple(),
        transports: [new winston.transports.Console()]
      }, logger));
    }
  }

  log(logObj) {
    const args = [].concat(logObj.args);
    const arg0 = args.shift();
    this.logger.log({
      level: levels[logObj.level] || 'info',
      label: logObj.tag,
      message: arg0,
      args: args,
      timestamp: logObj.date.getTime() / 1000
    });
  }

}
const levels = {
  0: 'error',
  1: 'warn',
  2: 'info',
  3: 'verbose',
  4: 'debug',
  5: 'silly'
};

exports.Consola = Consola;
exports.Types = Types;
exports.isLogObj = isLogObj;
exports.assignGlobalConsola = assignGlobalConsola;
exports.BasicReporter = BasicReporter;
exports.BrowserReporter = BrowserReporter;
exports.FancyReporter = FancyReporter;
exports.JSONReporter = JSONReporter;
exports.WinstonReporter = WinstonReporter;
